/************************************************************************/
/*  mssync:								*/
/*	Synchronize time corrections from a master MiniSEED file to	*/
/*	to another channel of data.					*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2002 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: mssync.c,v 1.4 2002/11/22 23:26:51 doug Exp $ ";
#endif

#include <stdio.h>
#include "version.h"

char *syntax[] = {
"%s version " VERSION ,
"%s - sync time corrections from ref MiniSEED file to other MiniSEED files .",
"Syntax:",
"%s    -r ref_file [-U] [-d n] [-h] [input_file [output_file]] | files...",
"    where:",
"	-r ref_file",
"		    Name of MiniSEED master file with time corrections.",
"	-U	    Update input file instead of creating new output file.",
"		    If you use this option, the file list is a list of",
"		    input/output files instead of an optional input and",
"		    output file.",
"	-d debug    Value of debug flag (may be ORed together)",
"		    1 = DEBUG_REF",
"	-h	    Help - prints syntax message.",
"	input_file",
"		    Name of the input file to be corrected.",
"		    If no file is specified, read from stdin.",
"	output_file",
"		    Name of the output file to be created.",
"		    If no file is specified, write to stdout.",
"	files...    A list of files to be synced and rewritten.",
"		    Used with the -U option.",
"Examples:",
"1.  Resync the BKS BHN file to BKS BHZ channel and write to new file.",
"    %s -r BKS.BK.BHZ.D.1997.200 BKS.BK.BHN.D.1997.200 BKS.BK.BHN.D.1997.200.new",
"2.  Resync all BKS HH? files to BKS BHZ channel and update the original files.",
"    %s -U r BKS.BK.BHZ.D.1997.200 BKS.BK.HH?.D.1997.200",
NULL };

#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#ifdef	SUNOS4
char *basename (const char *name);
char *dirname  (char *path);
#else
#include <libgen.h>
#endif

#include <qlib2.h>

#define	MAXLINELEN	256
#define ERROR		-2
#define MAXBLKSIZE	32768
#define	FIXED_DATA_HDR_SIZE 48
#define	DEBUG(x)	(debug_flag&(x))
#define	DEBUG_REF	1

char *cmdname;
char *ref_file = NULL;
char *in_file = NULL;
char *out_file = NULL;
FILE *info;
int date_fmt = JULIANC_FMT;
int update_flag = 0;
int debug_flag = 0;

INT_TIME basetime (DATA_HDR *hdr);
void print_time (INT_TIME it, int corr_cur, INT_TIME it1, 
		 int corr_1, INT_TIME it2, int corr_2);
void reset_ref_file (FILE *fp_ref);
int get_ref_hdr (INT_TIME it, DATA_HDR **p_ref_hdr1, DATA_HDR **p_ref_hdr2, 
		 FILE *fp_ref);

/************************************************************************/
/* main program.							*/
/************************************************************************/
main (int argc, char **argv) 
{
    FILE *fp_ref;
    FILE *fp_in = stdin;
    FILE *fp_out = stdout;
    int status;

    /* Variables needed for getopt. */
    extern char	*optarg;
    extern int	optind, opterr;
    int		c;

    /*	Parse command line options.					*/
    cmdname = tail(argv[0]);
    info = stdout;
    while ( (c = getopt(argc,argv,"hUd:r:")) != -1)
	switch (c) {
	case '?':
	case 'h':   print_syntax(cmdname,syntax,info); exit(0); break;
	case 'U':   update_flag = 1; break;
	case 'd':   debug_flag = atoi(optarg); break;
	case 'r':   ref_file = optarg; break;
	default:    fprintf (stderr, "Invalid option: -%c\n",c);
		    exit(1);
	}

    /*	Skip over all options and their arguments.			*/
    argv = &(argv[optind]);
    argc -= optind;

    if (! ref_file) {
	fprintf (stderr, "Error - must specify ref_file\n");
	print_syntax(cmdname,syntax,info);
	exit(1);
    }

    if (ref_file) {
	if ((fp_ref = fopen (ref_file, "r")) == NULL) {
	    fprintf (stderr, "Unable to open ref file %s\n", ref_file);
	    exit(1);
	}
    }

    /* Parse filelist if update_flag is not set.			*/
    if (!update_flag) {
	in_file = "<stdin>";
	if (--argc >= 0) {
	    in_file = *argv++;
	    if ((fp_in = fopen (in_file, "r")) == NULL) {
		fprintf (stderr, "Unable to open input file %s\n", in_file);
		exit(1);
	    }
	}

	out_file = "<stdout>";
	if (--argc >= 0) {
	    out_file = *argv++;
	    if ((fp_out = fopen (out_file, "w")) == NULL) {
		fprintf (stderr, "Unable to open output file %s\n", out_file);
		exit(1);
	    }
	}
	else info = stderr;
	status = sync_file (fp_in, fp_out, fp_ref);
    }

    else {
	/* Process each input file */
	char *pathcopy, *dir;
	while (argc-- > 0) {
	    reset_ref_file (fp_ref);
	    in_file = *argv++;
	    if ((fp_in = fopen (in_file, "r")) == NULL) {
		fprintf (stderr, "Unable to open input file %s\n", in_file);
		exit(1);
	    }
	    pathcopy = strdup(in_file);
	    dir = dirname (pathcopy);
	    out_file = tempnam (dir, cmdname);
	    free (pathcopy);
	    if ((fp_out = fopen (out_file, "w")) == NULL) {
		fprintf (stderr, "Unable to open output file %s\n", out_file);
		exit(1);
	    }
	    status = sync_file (fp_in, fp_out, fp_ref);
	    if (status != 0) {
		fprintf (stderr, "Error syncing file %s\n", in_file);
		exit (status);
	    }
	    fclose (fp_in);
	    fclose (fp_out);
	    status = rename (out_file, in_file);
	    if (status != 0) {
		fprintf (stderr, "Error renaming output file %s to %s - %s\n",
			 out_file, in_file, strerror (errno));
		exit (1);
	    }
	}
    }
    return (0);
}

/************************************************************************/
/*  sync_file:								*/
/*	Synchronize input file to reference file, output synced file.	*/
/************************************************************************/
int sync_file (FILE *fp_in, FILE *fp_out, FILE *fp_ref)
{
    char line[MAXLINELEN];
    char buf[MAXBLKSIZE];
    char out_buf[MAXBLKSIZE];
    char *p_buf;
    DATA_HDR *hdr, *ref_hdr1, *ref_hdr2;
    INT_TIME it, it1, it2, begtime;
    int corr_cur, corr_1, corr_2;
    int data_offset, datalen;
    double d1, d2;
    int status;

    hdr = NULL;
    ref_hdr1 = ref_hdr2 = NULL;
    /********************************************************************/
    /* loop over records in the input file.				*/
    /********************************************************************/
    corr_cur = 0;
    while (1) {
        p_buf = buf;
	status = read_ms_record (&hdr, &p_buf, fp_in);
	if (status < 0) break;

	it = basetime (hdr);

	/* Ensure that we have reference records that surround	*/
	/* the hdrtime of this record if possible.			*/
	get_ref_hdr (it, &ref_hdr1, &ref_hdr2, fp_ref);
	if (ref_hdr1 == NULL) {
	    /* No reference data.  Stay with current correction.	*/
	    it2 = it1;
	    corr_1 = corr_2 = corr_cur;
	}
	else {
	    /* Remove time correction if already applied to hdr.	*/
	    it1 = basetime (ref_hdr1);
	    corr_1 = ref_hdr1->num_ticks_correction;
	    if (ref_hdr2 != NULL) {
		it2 = basetime (ref_hdr2);
		corr_2 = ref_hdr2->num_ticks_correction;
	    }
	    else {
		it2 = it1;
		corr_2 = corr_1;
	    }
	}
	
	/* Figure out which correction to use.	*/
	d1 = d2 = 0;
	if (ref_hdr1 != NULL) {
	    d1 = fabs (tdiff (it, it1));
	    d2 = fabs (tdiff (it, it2));
	}
	corr_cur = (d1 <= d2) ? corr_1 : corr_2;

	if (DEBUG(DEBUG_REF)) {
	    print_time (it, corr_cur, it1, corr_1, it2, corr_2);
	}
	    
	/* Add correction to header. */
	begtime = add_time (it, 0, corr_cur);
	hdr->num_ticks_correction = corr_cur;
	hdr->begtime = add_time (it, 0, corr_cur);
	if (hdr->activity_flags & ACTIVITY_TIME_CORR_APPLIED) {
	    hdr->hdrtime = hdr->begtime;
	}
	else {
	    hdr->hdrtime = it;
	}

	/* Write the new record to the output. */
	data_offset = hdr->first_data;
	datalen = hdr->blksize - data_offset;
	memset ((void *)out_buf, 0, hdr->blksize);
	init_sdr_hdr ((SDR_HDR *)out_buf, hdr, NULL);
	if (datalen > 0) {
	    memcpy (out_buf+data_offset, buf+data_offset, datalen);
	}
	update_sdr_hdr ((SDR_HDR *)out_buf, hdr);
	status = fwrite (out_buf, hdr->blksize, 1, fp_out);
	if (status == 0) {
	    fprintf (stderr, "Error writing file %s\n", out_file);
	    exit (1);
	}
	free_data_hdr (hdr);
    }
    return (0);
}
	
/************************************************************************/
/*  basetime:								*/
/*	Return hdrtime with any time corrections removed.		*/
/************************************************************************/
INT_TIME basetime (DATA_HDR *hdr)
{
    INT_TIME it;
    it = hdr->hdrtime;
    if (hdr->activity_flags & ACTIVITY_TIME_CORR_APPLIED) {
	it = add_time (it, 0, -1*hdr->num_ticks_correction);
    }
    return (it);
}

/************************************************************************/
/*  print_time:								*/
/*	Print time and corrections for output_hdr and both ref_hdrs.	*/
/*	Used only for debugging.					*.
/************************************************************************/
void print_time (INT_TIME it, int corr_cur, INT_TIME it1, int corr_1, INT_TIME it2, int corr_2)
{
    fprintf (info,
	     "data = %s, %d ",
	     time_to_str(it, date_fmt), corr_cur);
    fprintf (info,
	     "ref1 = %s, %d ",
	     time_to_str(it1, date_fmt), corr_1);
    fprintf (info,
	     "ref2 = %s, %d\n",
	     time_to_str(it2, date_fmt), corr_2);
}

/************************************************************************/
/*  Variables for the reference file.					*/
/************************************************************************/
static DATA_HDR *ref_hdr1 = NULL;
static DATA_HDR *ref_hdr2 = NULL;
int eof_ref = 0;

/************************************************************************/
/*  reset_ref_file:							*/
/*	Reset the reference file to the initial condition.		*/
/************************************************************************/
void reset_ref_file (FILE *fp_ref)
{
    rewind (fp_ref);
    eof_ref = 0;
    ref_hdr1 = ref_hdr2 = NULL;
}

/************************************************************************/
/*  get_ref_hdr:							*/
/*	Return ptrs (if possible) to the record headers from the	*/
/*	reference MiniSEED file whose hdrtimes WITHOUT corrections	*/
/*	bracket the specified time.					*/
/************************************************************************/
int get_ref_hdr (INT_TIME it, DATA_HDR **p_ref_hdr1, DATA_HDR **p_ref_hdr2, FILE *fp_ref)
{
    INT_TIME it1;
    INT_TIME it2;
    int status;
    char buf[MAXBLKSIZE];

    *p_ref_hdr1 = ref_hdr1;
    *p_ref_hdr2 = ref_hdr2;
    if (eof_ref) return (0);

    /* Get first header if we don't have one already */
    if (ref_hdr1 == NULL) {
	status = read_ms_record (&ref_hdr1, (char **)&buf, fp_ref);
	if (status < 0) {
	    if (status == EOF) {
		eof_ref = 1;
		return (0);
	    }
	    fprintf (stderr, "Error reading reference file %s\n",
		     ref_file);
	    exit (1);
	}
    }
    it1 = basetime (ref_hdr1);

    if (ref_hdr2 == NULL) {
	status = read_ms_record (&ref_hdr2, (char **)&buf, fp_ref);
	if (status < 0) {
	    if (status == EOF) {
		eof_ref = 1;
		return (0);
	    }
	    fprintf (stderr, "Error reading reference file %s\n",
		     ref_file);
	    exit (1);
	}
    }
    it2 = basetime (ref_hdr2);

    /* Keep reading records until we get a record whose base time is	*/
    /* greater than or equal to the specified base time, or we see EOF.	*/
    while (tdiff(it,it2) > 0) {
	free_data_hdr (ref_hdr1);
	ref_hdr1 = ref_hdr2;
	it1 = it2;
	ref_hdr2 = NULL;
	status = read_ms_record (&ref_hdr2, (char **)&buf, fp_ref);
	if (status < 0) {
	    if (status == EOF) {
		eof_ref = 1;
		break;
	    }
	    fprintf (stderr, "Error reading reference file %s\n",
		     ref_file);
	    exit (1);
	}
	it2 = basetime (ref_hdr2);
    }

    *p_ref_hdr1 = ref_hdr1;
    *p_ref_hdr2 = ref_hdr2;
    return (0);
}

